/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
<pre>{@code
                   y
                   |
                   |
                   | v[4]
                 1 +-----------------+ v[5]=(1,1,0)
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |                    y
           |       |         |       |                    |
           |       |         |       | v[1]               |
           |  v[0] +---------|-------+------> x           |
           |      /          |      /1                    |
           |    /            |    /                       +-----> x
           |  /              |  /                        /
           |/                |/                         /
         1 +-----------------+                         /
         / v[3]=(0,0,1)      v[2]=(1,0,1)            z
       /
     /
    z
}</pre>
   Render a single wireframe cube.
*/
@SuppressWarnings("serial")
public class InteractiveCube_R3 extends InteractiveFrame
{
   private Scene scene;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public InteractiveCube_R3(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create a Model object to hold the geometry.
      Model model = new Model();

      // Create the geometry for the Model.
      Vertex v0, v1, v2, v3, v4, v5, v6, v7;
      v0 = new Vertex(0.0, 0.0, 0.0); // four vertices around the bottom face
      v1 = new Vertex(1.0, 0.0, 0.0);
      v2 = new Vertex(1.0, 0.0, 1.0);
      v3 = new Vertex(0.0, 0.0, 1.0);
      v4 = new Vertex(0.0, 1.0, 0.0); // four vertices around the top face
      v5 = new Vertex(1.0, 1.0, 0.0);
      v6 = new Vertex(1.0, 1.0, 1.0);
      v7 = new Vertex(0.0, 1.0, 1.0);

      // Add the geometry to the Model.
      model.addLineSegment(v0, v1);  // bottom face
      model.addLineSegment(v1, v2);  // bottom face
      model.addLineSegment(v2, v3);  // bottom face
      model.addLineSegment(v3, v0);  // bottom face
      model.addLineSegment(v4, v5);  // top face
      model.addLineSegment(v5, v6);  // top face
      model.addLineSegment(v6, v7);  // top face
      model.addLineSegment(v7, v4);  // top face
      model.addLineSegment(v0, v4);  // back face
      model.addLineSegment(v1, v5);  // back face
      model.addLineSegment(v3, v7);  // front face
      model.addLineSegment(v2, v6);  // front face
/*
      // Or we can add the geometry to the model in this more efficient way.
      model.addVertex(v0, v1, v2, v3, v4, v5, v6, v7);
      model.addLineSegment(0, 1);  // bottom face
      model.addLineSegment(1, 2);  // bottom face
      model.addLineSegment(2, 3);  // bottom face
      model.addLineSegment(3, 0);  // bottom face
      model.addLineSegment(4, 5);  // top face
      model.addLineSegment(5, 6);  // top face
      model.addLineSegment(6, 7);  // top face
      model.addLineSegment(7, 4);  // top face
      model.addLineSegment(0, 4);  // back face
      model.addLineSegment(1, 5);  // back face
      model.addLineSegment(3, 7);  // front face
      model.addLineSegment(2, 6);  // front face
*/
      // Push the model away from where the camera is.
      for (Vertex v : model.vertexList)
      {
         v.z -= 2;
      }


      // Create the Scene object that we shall render.
      scene = new Scene();

      // Add the Model to the Scene.
      scene.addModel(model);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
         Clip.debug = ! Clip.debug;
       //Rasterize.debug = ! Rasterize.debug;
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
      }
      else if ('s' == c) // Scale the cube 10% smaller.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x /= 1.1;  // This does NOT work with
            v.y /= 1.1;  // perspective projection.
            v.z /= 1.1;
         }
      }
      else if ('S' == c) // Scale the cube 10% larger.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x *= 1.1;  // This does NOT work with
            v.y *= 1.1;  // perspective projection.
            v.z *= 1.1;
         }
      }
      else if ('x' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x -= 0.1;
         }
      }
      else if ('X' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x += 0.1;
         }
      }
      else if ('y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y -= 0.1;
         }
      }
      else if ('Y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y += 0.1;
         }
      }
      else if ('z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z -= 0.1;
         }
      }
      else if ('Z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z += 0.1;
         }
      }

      // Render again.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // InteractiveCube_R3 constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new InteractiveCube_R3("Renderer 3", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the cube along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the cube.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
